# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

import datetime
# garbage collector for memory issue
import gc
import os

import numpy as np
from numpy.lib import recfunctions as rfn

from remotior_sensus.core import configurations as cfg, table_manager as tm
from remotior_sensus.core.log import Log
from remotior_sensus.util import files_directories, raster_vector


def function_initiator(
        process_parameters=None, input_parameters=None, output_parameters=None,
        function=None, function_argument=None, function_variable=None,
        run_separate_process=False, classification=False,
        classification_confidence=False, signature_raster=False
):
    # get process parameters
    process_id = str(process_parameters[0])
    cfg.temp = process_parameters[1]
    memory = process_parameters[2]
    gdal_path = process_parameters[3]
    progress_queue = process_parameters[4]
    refresh_time = process_parameters[5]
    # get input raster parameters
    raster_list = input_parameters[0]
    calc_data_type = input_parameters[1]
    boundary_size = input_parameters[2]
    if input_parameters[3] is None:
        x_min_piece = 0
        y_min_piece = 0
        x_size_piece = y_size_piece = sections = None
    else:
        x_min_piece = input_parameters[3].x_min
        x_size_piece = input_parameters[3].x_size
        sections = input_parameters[3].sections
        y_min_piece_no_boundary = input_parameters[3].y_min_no_boundary
        if y_min_piece_no_boundary is not None:
            y_min_piece = y_min_piece_no_boundary
        else:
            y_min_piece = input_parameters[3].y_min
        y_size_piece_no_boundary = input_parameters[3].y_size_no_boundary
        if y_size_piece_no_boundary is not None:
            y_size_piece = y_size_piece_no_boundary
        else:
            y_size_piece = input_parameters[3].y_size
    scale = input_parameters[4]
    offset = input_parameters[5]
    use_value_as_nodata = input_parameters[6]
    single_band_number = input_parameters[7]
    input_nodata_as_value = input_parameters[8]
    multi_add_factors = input_parameters[9]
    dummy_bands = input_parameters[10]
    specific_output = input_parameters[11]
    if specific_output is not None:
        specific_output_piece = specific_output['pieces'][int(process_id)]
    else:
        specific_output_piece = None
    # get output parameters
    (output_raster_list, output_data_type, compress, compress_format,
     any_nodata_mask, output_no_data, output_band_number, keep_output_array,
     keep_output_argument) = output_parameters
    # start logger
    cfg.logger = Log(directory=cfg.temp.dir, multiprocess=str(process_id))
    # set gdal path
    if gdal_path is not None:
        for d in gdal_path.split(';'):
            try:
                os.add_dll_directory(d)
                cfg.gdal_path = d
            except Exception as err:
                str(err)
    # import gdal
    from osgeo import gdal
    # GDAL config
    try:
        gdal.SetConfigOption('GDAL_DISABLE_READDIR_ON_OPEN', 'TRUE')
        gdal.SetConfigOption('GDAL_CACHEMAX', str(memory))
        gdal.SetConfigOption('VSI_CACHE', 'FALSE')
        gdal.SetConfigOption('CHECK_DISK_FREE_SPACE', 'FALSE')
    except Exception as err:
        str(err)
    cfg.logger.log.debug('start')
    cfg.logger.log.debug(
        'keep_output_array: %s; keep_output_argument: %s; any_nodata_mask: %s;'
        ' input_nodata_as_value: %s'
        % (keep_output_array, keep_output_argument, any_nodata_mask,
           input_nodata_as_value)
    )
    # list of output arrays
    output_array_list = []
    # output files
    out_files = []
    # output classification
    output_signature_raster = {}
    output_signature_raster_dic = {}
    out_class = out_alg = None
    # process error
    proc_error = False
    # raster counter
    raster_count = 0
    start_time = datetime.datetime.now()
    cfg.logger.log.debug(
        'raster_list: %s; x_size_piece: %s; y_size_piece: %s:'
        % (raster_list, x_size_piece, y_size_piece)
    )
    # iterate over input raster list
    for raster in raster_list:
        # reset section counter
        count_section_progress = 1
        calculation_datatype = calc_data_type[raster_count]
        cfg.logger.log.debug(
            'process raster: %s; calculation_datatype: %s'
            % (raster, calculation_datatype)
        )
        if use_value_as_nodata:
            value_as_nodata = use_value_as_nodata[raster_count]
        else:
            value_as_nodata = None
        (r_gt, r_crs, r_un, r_xy_count, r_nd, band_number, r_block_size,
         r_scale_offset, r_data_type) = raster_vector.raster_info(raster)
        cfg.logger.log.debug('r_gt: {}'.format(str(r_gt)))
        # pixel size and origin from reference
        t_lx = r_gt[0]
        t_ly = r_gt[3]
        p_sx = r_gt[1]
        p_sy = r_gt[5]
        # scale and offset
        if scale:
            if scale[raster_count] is not None:
                scl = float(scale[raster_count])
            else:
                scl = 1.0
        else:
            scl = 1.0
        if offset:
            if offset[raster_count] is not None:
                offs = float(offset[raster_count])
            else:
                offs = 0.0
        else:
            offs = 0.0
        # calculate sections for separate process
        if run_separate_process:
            # raster size
            r_x, r_y = r_xy_count
            # raster blocks
            memory_unit = cfg.memory_unit_array_12
            (block_size_x, block_size_y,
             list_range_x, list_range_y, tot_blocks) = _calculate_block_size(
                x_block=r_x, y_block=r_y, available_ram=memory,
                memory_unit=memory_unit, dummy_bands=dummy_bands
            )
            cfg.logger.log.debug('list_range_y: %s' % str(list_range_y))
            sections = []
            for y_min in range(len(list_range_y)):
                # section y_min size
                s_y_size = block_size_y
                # adapt to raster y_min size
                if list_range_y[y_min] + s_y_size > r_y:
                    s_y_size = r_y - list_range_y[y_min]
                sections.append(
                    RasterSection(
                        x_min=0, y_min=list_range_y[y_min], x_max=block_size_x,
                        y_max=list_range_y[y_min] + s_y_size,
                        x_size=block_size_x, y_size=s_y_size
                    )
                )
        cfg.logger.log.debug('sections: %s' % str(sections))
        # classification rasters
        classification_rasters = []
        # algorithm rasters
        algorithm_rasters = []
        # set raster parameters
        output_argument = None
        # generic output raster
        out_generic = None
        # specific output raster
        out_specific = None
        # output data type
        if output_data_type:
            out_data_type = output_data_type[raster_count]
        else:
            out_data_type = None
        # output nodata
        if output_no_data:
            out_no_data = output_no_data[raster_count]
        else:
            out_no_data = None
        # create output rasters if process with output files
        if output_raster_list[0] is not None:
            # output geo transform
            upper_left_x_range = t_lx + x_min_piece * p_sx
            upper_left_y_range = t_ly + y_min_piece * p_sy
            geo_transform = (upper_left_x_range, p_sx, r_gt[2],
                             upper_left_y_range, r_gt[4], p_sy)
            # classification output
            if classification:
                signatures = function_argument[0][
                    cfg.spectral_signatures_framework]
                # get selected signatures
                signatures_table = signatures.table[
                    signatures.table.selected == 1]
                # list of rasters to be created
                output_classification_raster_list = []
                output_signature_raster = {}
                output_signature_raster_dic = {}
                # create output raster paths
                out_class = cfg.temp.temporary_raster_path(
                    name='class_', name_suffix=process_id
                )
                output_classification_raster_list.append(out_class)
                classification_rasters.append(out_class)
                # signature output rasters
                if signature_raster:
                    for s in signatures_table.signature_id.tolist():
                        sig_name = '{}{}_{}_'.format(
                            s, signatures_table[
                                signatures_table.signature_id ==
                                s].macroclass_id[0],
                            signatures_table[
                                signatures_table.signature_id == s].class_id[0]
                        )
                        signature_file_path = cfg.temp.temporary_raster_path(
                            name=sig_name, name_suffix=process_id
                        )
                        output_signature_raster[s] = signature_file_path
                        if s in output_signature_raster_dic:
                            output_signature_raster_dic[s].append(
                                signature_file_path
                            )
                        else:
                            output_signature_raster_dic[
                                s] = [signature_file_path]
                        output_classification_raster_list.append(
                            signature_file_path
                        )
                # confidence raster
                if classification_confidence:
                    out_alg = cfg.temp.temporary_raster_path(
                        name='alg_', name_suffix=process_id
                    )
                    output_classification_raster_list.append(out_alg)
                    algorithm_rasters.append(out_alg)
                # create rasters
                raster_vector.create_raster_from_reference(
                    path=raster, band_number=1,
                    output_raster_list=output_classification_raster_list,
                    nodata_value=out_no_data, driver='GTiff',
                    gdal_format=out_data_type, compress=compress,
                    compress_format=compress_format,
                    geo_transform=geo_transform, x_size=x_size_piece,
                    y_size=y_size_piece
                )
                cfg.logger.log.debug(
                    'output_classification_raster_list: %s'
                    % str(output_classification_raster_list)
                )
            # specific output
            elif specific_output is not None:
                x_min_piece_s = specific_output_piece.x_min
                y_min_piece_s = specific_output_piece.y_min
                x_size_piece_s = specific_output_piece.x_size
                y_size_piece_s = specific_output_piece.y_size
                # output geo transform
                upper_left_x_range = (specific_output['geo_transform'][0]
                                      + x_min_piece_s
                                      * specific_output['geo_transform'][1])
                upper_left_y_range = (specific_output['geo_transform'][3]
                                      + y_min_piece_s
                                      * specific_output['geo_transform'][5])
                geo_transform = (upper_left_x_range,
                                 specific_output['geo_transform'][1],
                                 specific_output['geo_transform'][2],
                                 upper_left_y_range,
                                 specific_output['geo_transform'][4],
                                 specific_output['geo_transform'][5])
                out_specific = cfg.temp.temporary_raster_path(
                    name_suffix=process_id, name_prefix=process_id
                )
                file_output = raster_vector.create_raster_from_reference(
                    raster, 1, [out_specific], out_no_data, 'GTiff',
                    out_data_type, compress, compress_format,
                    geo_transform=geo_transform, x_size=x_size_piece_s,
                    y_size=y_size_piece_s
                )
                cfg.logger.log.debug('specific_output: %s' % file_output)
            # generic output
            else:
                out_generic = cfg.temp.temporary_raster_path(
                    name_suffix=process_id, name_prefix=process_id
                )
                file_output = raster_vector.create_raster_from_reference(
                    raster, 1, [out_generic], out_no_data, 'GTiff',
                    out_data_type, compress, compress_format,
                    geo_transform=geo_transform, x_size=x_size_piece,
                    y_size=y_size_piece
                )
                cfg.logger.log.debug('file_output: %s' % file_output)
        # iterate over sections
        for sec in sections:
            # open input_raster with GDAL
            _r_d = gdal.Open(raster, gdal.GA_ReadOnly)
            cfg.logger.log.debug('raster: %s' % raster)
            # list of bands
            gdal_band_list = []
            if single_band_number is None:
                for b in range(1, band_number + 1):
                    r_b = _r_d.GetRasterBand(b)
                    gdal_band_list.append(r_b)
            else:
                r_b = _r_d.GetRasterBand(single_band_number + 1)
                gdal_band_list.append(r_b)
            _input_array = np.zeros(
                (sec.y_size, sec.x_size, len(gdal_band_list)),
                dtype=calculation_datatype
            )
            nodata_mask = None
            nd_val = None
            # read bands
            for b in range(len(gdal_band_list)):
                cfg.logger.log.debug(
                    'sec.x_min: %s; sec.y_min: %s; sec.x_size: %s; '
                    'sec.y_size: %s'
                    % (sec.x_min, sec.y_min, sec.x_size, sec.y_size)
                )
                # band array
                _a = raster_vector.read_array_block(
                    gdal_band_list[b], sec.x_min, sec.y_min, sec.x_size,
                    sec.y_size, calculation_datatype
                )
                # get band nodata, scale and offset
                try:
                    # if single band
                    band_b = gdal_band_list[b].GetRasterBand(1)
                    offs_b = band_b.GetOffset()
                    scl_b = band_b.GetScale()
                    if scl_b is not None:
                        scl_b = scl_b
                    if offs_b is not None:
                        offs_b = offs_b
                    offs_b = np.asarray(offs_b).astype(_a.dtype)
                    scl_b = np.asarray(scl_b).astype(_a.dtype)
                    nd_val = band_b.GetNoDataValue()
                    ndv_band = np.asarray(
                        nd_val * scl_b + offs_b
                    ).astype(_a.dtype)
                except Exception as err:
                    str(err)
                    try:
                        offs_b = gdal_band_list[b].GetOffset()
                        scl_b = gdal_band_list[b].GetScale()
                        if scl_b is not None:
                            scl_b = scl_b
                        else:
                            scl_b = 1
                        if offs_b is not None:
                            offs_b = offs_b
                        else:
                            offs_b = 0
                        offs_b = np.asarray(offs_b).astype(_a.dtype)
                        scl_b = np.asarray(scl_b).astype(_a.dtype)
                        nd_val = gdal_band_list[b].GetNoDataValue()
                        ndv_band = np.asarray(
                            nd_val * scl_b + offs_b
                        ).astype(_a.dtype)
                    except Exception as err:
                        str(err)
                        ndv_band = None
                if ndv_band is not None:
                    # adapt NoData to dtype
                    ndv_band = np.asarray(ndv_band).astype(
                        calculation_datatype
                    )
                if _a is None:
                    proc_error = 'input raster'
                else:
                    # apply multiplicative and additive factors to array
                    if multi_add_factors is not None:
                        _a = array_multiplicative_additive_factors(
                            _a, multi_add_factors[0][b],
                            multi_add_factors[1][b]
                        )
                    _input_array[::, ::, b] = _a.astype(calculation_datatype)
                # set nodata value
                if calculation_datatype == np.float32 or \
                        calculation_datatype == np.float64:
                    if ndv_band is not None:
                        try:
                            _input_array[::, ::, b][
                                _input_array[::, ::, b] == ndv_band] = np.nan
                        except Exception as err:
                            str(err)
                    if input_nodata_as_value:
                        try:
                            _input_array[::, ::, b][
                                np.isnan(_input_array[::, ::, b])] = ndv_band
                        except Exception as err:
                            str(err)
                cfg.logger.log.debug(
                    'gdal_band_list[%s].DataType: %s; nd_val: %s; ndv_band: %s'
                    % (b, gdal_band_list[b].DataType, nd_val, ndv_band)
                )
                # apply NoData mask
                if any_nodata_mask:
                    try:
                        nodata_mask[0, 0]
                    except Exception as err:
                        str(err)
                        nodata_mask = np.zeros(
                            (sec.y_size, sec.x_size),
                            dtype=calculation_datatype
                        )
                    if input_nodata_as_value:
                        pass
                    elif ndv_band is not None:
                        try:
                            np.copyto(
                                nodata_mask, out_no_data,
                                where=_input_array[::, ::, b] == np.asarray(
                                    ndv_band
                                )
                            )
                        except Exception as err:
                            str(err)
                    if input_nodata_as_value:
                        pass
                    else:
                        try:
                            np.copyto(
                                nodata_mask, out_no_data,
                                where=np.isnan(_input_array[::, ::, b])
                            )
                        except Exception as err:
                            str(err)
                    if input_nodata_as_value:
                        pass
                    elif value_as_nodata is not None:
                        value_as_nodata = np.asarray(value_as_nodata).astype(
                            _a.dtype
                        ).astype(calculation_datatype)
                        try:
                            np.copyto(
                                nodata_mask, out_no_data,
                                where=_input_array[
                                      ::, ::, b] == value_as_nodata
                            )
                        except Exception as err:
                            str(err)
                # create NoData mask from NoData values
                elif not any_nodata_mask:
                    _all_nodata_mask = np.zeros(
                        (sec.y_size, sec.x_size), dtype=calculation_datatype
                    )
                    try:
                        nodata_mask[0, 0]
                    except Exception as err:
                        str(err)
                        nodata_mask = np.ones(
                            (sec.y_size, sec.x_size),
                            dtype=calculation_datatype
                        ) * out_no_data
                    if input_nodata_as_value:
                        pass
                    elif ndv_band is not None:
                        try:
                            np.copyto(
                                _all_nodata_mask, out_no_data,
                                where=_input_array[::, ::, b] == ndv_band
                            )
                        except Exception as err:
                            str(err)
                    if input_nodata_as_value:
                        pass
                    else:
                        try:
                            np.copyto(
                                _all_nodata_mask, out_no_data,
                                where=np.isnan(_input_array[::, ::, b])
                            )
                        except Exception as err:
                            str(err)
                    if input_nodata_as_value:
                        pass
                    elif value_as_nodata is not None:
                        value_as_nodata = np.asarray(value_as_nodata).astype(
                            _a.dtype
                        ).astype(calculation_datatype)
                        try:
                            np.copyto(
                                _all_nodata_mask, out_no_data,
                                where=_input_array[
                                      ::, ::, b] == value_as_nodata
                            )
                        except Exception as err:
                            str(err)
                    nodata_mask = np.where(
                        (_all_nodata_mask == out_no_data) & (
                                nodata_mask == out_no_data),
                        out_no_data, 0
                    )
                    # release memory
                    _all_nodata_mask = None
            # release memory
            _a = None
            # get function variable
            if function_variable:
                f_variable = function_variable[raster_count]
            else:
                f_variable = None
            # get output raster
            if output_raster_list[0]:
                output_list = output_raster_list[raster_count]
            else:
                output_list = None
            # get output band number
            if output_band_number:
                out_band_number = output_band_number[raster_count]
            else:
                out_band_number = None
            # get function argument
            if function_argument:
                function_arg = function_argument[raster_count]
            else:
                function_arg = None
            # execute main function
            function_output = function(
                [scl, offs, out_no_data], _input_array, nodata_mask,
                sec.y_size, sec.x_min, sec.y_min, output_list,
                function_arg, f_variable, out_band_number,
                [x_min_piece, y_min_piece], output_signature_raster,
                out_class, out_alg
            )
            # release memory
            _input_array = None
            # check function output list
            if isinstance(function_output, list):
                output_argument = function_output[1]
                output_array = function_output[0]
            else:
                proc_error = 'error function'
                output_array = None
            # classification
            if classification:
                output_array = None
                output_argument = None
                output_array_list.append(
                    [classification_rasters, algorithm_rasters,
                     output_signature_raster_dic]
                )
                cfg.logger.log.debug('classification: %s' % function_output[0])
                if not function_output[0] and function_output[0] is not None:
                    proc_error = 'error classification'
                    cfg.logger.log.error('classification failed')
            # keep output array and output argument
            elif keep_output_array and keep_output_array is not None and (
                    keep_output_argument is True and keep_output_argument is
                    not None):
                output_array_list.append([output_array, output_argument])
            # keep output array
            elif keep_output_array and keep_output_array is not None:
                output_array_list.append([output_array, None])
            # keep output argument
            elif keep_output_argument and keep_output_argument is not None:
                output_array_list.append([None, output_argument])
            # output files
            cfg.logger.log.debug('output_raster_list: %s' % output_raster_list)
            # write output array
            if (classification is not True
                    and output_raster_list[0] is not None):
                if specific_output is not None:

                    output_array[np.isnan(output_array)] = out_no_data

                    y_min_piece_s = specific_output_piece.y_min
                    if boundary_size is None:
                        # output minimum y (section of piece)
                        out_y_min = (specific_output_piece.sections[
                                        count_section_progress - 1].y_min
                                     - y_min_piece_s)
                    else:
                        # output minimum y (section of piece)
                        out_y_min = (
                                specific_output_piece.sections[
                                    count_section_progress
                                    - 1].y_min_no_boundary - y_min_piece_s)
                        # reduce array size without boundary
                        output_array = output_array[
                                       specific_output_piece.sections[
                                           count_section_progress
                                           - 1].y_size_boundary_top:(
                                               specific_output_piece.sections[
                                                   count_section_progress
                                                   - 1].y_size_boundary_top +
                                               specific_output_piece.sections[
                                                   count_section_progress
                                                   - 1].y_size_no_boundary),
                                       0:specific_output_piece.sections[
                                           count_section_progress - 1].x_max]
                    try:
                        write_out = raster_vector.write_raster(
                            out_specific, 0, out_y_min, output_array,
                            out_no_data, scl, offs
                        )
                        if write_out == out_specific:
                            if run_separate_process:
                                pass
                            else:
                                out_files.append([out_specific,
                                                  output_argument])
                    except Exception as err:
                        proc_error = 'error output'
                        cfg.logger.log.error(str(err))
                else:
                    try:
                        output_array[np.isnan(output_array)] = out_no_data
                        if boundary_size is None:
                            # output minimum y (section of piece)
                            out_y_min = sec.y_min - y_min_piece
                        else:
                            # output minimum y (section of piece)
                            out_y_min = sec.y_min_no_boundary - y_min_piece
                            # reduce array size without boundary
                            output_array = output_array[
                                           sec.y_size_boundary_top:(
                                                   sec.y_size_boundary_top
                                                   + sec.y_size_no_boundary),
                                           0:sec.x_max]
                        cfg.logger.log.debug('out_generic: %s' % out_generic)
                        write_out = raster_vector.write_raster(
                            out_generic, 0, out_y_min, output_array,
                            out_no_data, scl, offs
                        )
                        if write_out == out_generic:
                            if run_separate_process:
                                pass
                            else:
                                out_files.append([out_generic,
                                                  output_argument])
                    except Exception as err:
                        proc_error = 'error output'
                        cfg.logger.log.error(str(err))
            # progress
            now_time = datetime.datetime.now()
            elapsed_time = (now_time - start_time).total_seconds()
            if (process_id == '0' and elapsed_time > refresh_time
                    and not run_separate_process):
                start_time = now_time
                progress_queue.put(
                    [count_section_progress, len(sections)], False
                )
            elif (process_id == '1' and elapsed_time > refresh_time
                    and run_separate_process):
                start_time = now_time
                progress_queue.put(
                    [count_section_progress
                     + len(sections) * raster_count,
                     len(sections) * len(raster_list)], False
                )
            # close GDAL rasters
            for b in range(len(gdal_band_list)):
                gdal_band_list[b].FlushCache()
                gdal_band_list[b] = None
            _r_d = None
            gc.collect()
            count_section_progress += 1
        if classification is not True and output_raster_list[0] is not None:
            if run_separate_process:
                # move temporary files of separate process
                files_directories.move_file(
                    out_generic, output_raster_list[raster_count]
                )
                out_files.append(
                    [output_raster_list[raster_count],
                     output_argument]
                )
        raster_count += 1
        """
        now_time = datetime.datetime.now()
        elapsed_time = (now_time - start_time).total_seconds()
        if (process_id == '1' and elapsed_time > refresh_time
                and run_separate_process):
            progress_queue.put([raster_count, len(raster_list)], False)
        """
    cfg.logger.log.debug('end')
    logger = cfg.logger.stream.getvalue()
    return output_array_list, out_files, proc_error, logger


# calculate block size and pixel ranges
def _calculate_block_size(
        x_block, y_block, available_ram, memory_unit, dummy_bands
):
    single_block_size = x_block * y_block * memory_unit * (1 + dummy_bands)
    ram_blocks = int(single_block_size / available_ram)
    if ram_blocks == 0:
        ram_blocks = 1
    cfg.logger.log.debug('ram_blocks: %s' % str(ram_blocks))
    block_size_x = x_block
    list_range_x = list(range(0, x_block, block_size_x))
    block_size_y = int(y_block / ram_blocks)
    list_range_y = list(range(0, y_block, block_size_y))
    tot_blocks = len(list_range_x) * len(list_range_y)
    return (block_size_x, block_size_y, list_range_x, list_range_y,
            tot_blocks)


# table join
def table_join(
        table_1_parameters, table_2_parameters, nodata_value, join_type,
        features, output_names, process_parameters
):
    # get parameters
    (table1, field1_name, table1_names, table1_dtypes, table1_features,
     table1_features_index) = table_1_parameters
    (table2, field2_name, table2_names, table2_dtypes, table2_features,
     table2_features_index, table2_output_names,
     features_table_2_outer) = table_2_parameters
    process_id, cfg.temp, progress_queue, refresh_time = process_parameters
    proc_error = False
    cfg.logger = Log(directory=cfg.temp.dir, multiprocess=str(process_id))
    cfg.logger.log.debug('start')
    c = 0
    output = output_part = outer_part = None
    start_time = datetime.datetime.now()
    cfg.logger.log.debug('join_type: %s' % join_type)
    for feature in features:
        c += 1
        now_time = datetime.datetime.now()
        elapsed_time = (now_time - start_time).total_seconds()
        if process_id == 0 and elapsed_time > refresh_time:
            start_time = now_time
            progress_queue.put([c, len(features)], False)
        if join_type == 'left':
            # table 1 features
            output_part = table1[table1[field1_name] == feature]
            # table 2 row of unique features
            if feature in table2_features:
                row_table_2 = table2[table2_features_index[
                    np.where(table2_features == feature)]]
            else:
                row_table_2 = None
            # add table 2 fields and features
            for column_2 in range(len(table2_output_names)):
                if row_table_2 is None:
                    new_field = np.array(
                        [nodata_value] * output_part.shape[0],
                        dtype=table2_dtypes[column_2]
                    )
                else:
                    new_field = np.array(
                        list(row_table_2[table2_names[column_2]])
                        * output_part.shape[0],
                        dtype=table2_dtypes[column_2]
                    )
                output_part = tm.append_field(
                    output_part, table2_output_names[column_2],
                    new_field, table2_dtypes[column_2]
                )
        elif join_type == 'right':
            # table 2 features
            output_part = table2[table2[field2_name] == feature]
            # rename fields
            output_part = tm.redefine_matrix_columns(
                matrix=output_part, input_field_names=table2_names,
                output_field_names=table2_output_names,
                progress_message=False
            )
            # table 1 row of unique features
            if feature in table1_features:
                row_table_1 = table1[table1_features_index[
                    np.where(table1_features == feature)]]
            else:
                row_table_1 = None
            # add table 1 fields and features
            for column_1 in range(len(table1_names)):
                # get field1 name
                if table1_names[column_1] == field1_name:
                    new_field = np.array(
                        [feature] * output_part.shape[0],
                        dtype=table1_dtypes[column_1]
                    )
                elif row_table_1 is None:
                    new_field = np.array(
                        [nodata_value] * output_part.shape[0],
                        dtype=table1_dtypes[column_1]
                    )
                else:
                    new_field = np.array(
                        list(row_table_1[table1_names[column_1]])
                        * output_part.shape[0],
                        dtype=table1_dtypes[column_1]
                    )
                output_part = tm.append_field(
                    output_part, table1_names[column_1], new_field,
                    table1_dtypes[column_1]
                )
        elif join_type == 'inner':
            # table 1 features
            output_part = table1[table1[field1_name] == feature]
            # table 2 row of unique features
            row_table_2 = table2[table2_features_index[
                np.where(table2_features == feature)]]
            # add table 2 fields and features
            for column_2 in range(len(table2_output_names)):
                new_field = np.array(
                    list(row_table_2[table2_names[column_2]])
                    * output_part.shape[0], dtype=table2_dtypes[column_2]
                )
                output_part = tm.append_field(
                    output_part, table2_output_names[column_2],
                    new_field, table2_dtypes[column_2]
                )
        # full outer join
        elif join_type == 'outer':
            # table 1 features
            output_part = table1[table1[field1_name] == feature]
            # table 2 row of unique features
            if feature in table2_features:
                row_table_2 = table2[table2_features_index[
                    np.where(table2_features == feature)]]
            else:
                row_table_2 = None
            # add table 2 fields and features
            for column_2 in range(len(table2_output_names)):
                if row_table_2 is None:
                    new_field = np.array(
                        [nodata_value] * output_part.shape[0],
                        dtype=table2_dtypes[column_2]
                    )
                else:
                    new_field = np.array(
                        list(row_table_2[table2_names[column_2]])
                        * output_part.shape[0],
                        dtype=table2_dtypes[column_2]
                    )
                output_part = tm.append_field(
                    output_part, table2_output_names[column_2],
                    new_field, table2_dtypes[column_2]
                )
        if output is None:
            output = output_part
        else:
            output = rfn.stack_arrays(
                (output, output_part), asrecarray=True, usemask=False
            )
    # complete outer join with remaining features
    if (join_type == 'outer' and process_id == 0
            and len(features_table_2_outer) > 0):
        # iterate main features
        for feature in features_table_2_outer:
            # table 2 features
            outer_part = table2[table2[field2_name] == feature]
            # rename fields
            outer_part = tm.redefine_matrix_columns(
                matrix=outer_part, input_field_names=table2_names,
                output_field_names=table2_output_names,
                progress_message=False
            )
            # add table 1 fields and features
            for column_1 in range(len(table1_names)):
                # get field1 name
                if table1_names[column_1] == field1_name:
                    new_field = np.array(
                        [feature] * outer_part.shape[0],
                        dtype=table1_dtypes[column_1]
                    )
                else:
                    new_field = np.array(
                        [nodata_value] * outer_part.shape[0],
                        dtype=table1_dtypes[column_1]
                    )
                outer_part = tm.append_field(
                    outer_part, table1_names[column_1], new_field,
                    table1_dtypes[column_1]
                )
        if output is None:
            output = outer_part
        else:
            output = rfn.stack_arrays(
                (output, outer_part), asrecarray=True, usemask=False
            )
    # arrange fields
    output = tm.redefine_matrix_columns(
        matrix=output, input_field_names=output_names,
        output_field_names=output_names, progress_message=False
    )
    cfg.logger.log.debug('end')
    logger = cfg.logger.stream.getvalue()
    return output, None, proc_error, logger


# gdal translate processor
def gdal_translate(
        input_file=None, output=None, option_string=None,
        process_parameters=None
):
    process_id = process_parameters[0]
    cfg.temp = process_parameters[1]
    memory = process_parameters[2]
    gdal_path = process_parameters[3]
    progress_queue = process_parameters[4]
    cfg.logger = Log(directory=cfg.temp.dir, multiprocess='0')
    cfg.logger.log.debug('start')
    if gdal_path is not None:
        for d in gdal_path.split(';'):
            try:
                os.add_dll_directory(d)
                cfg.gdal_path = d
            except Exception as err:
                str(err)
    from osgeo import gdal
    # GDAL config
    try:
        gdal.SetConfigOption('GDAL_DISABLE_READDIR_ON_OPEN', 'TRUE')
        gdal.SetConfigOption('GDAL_CACHEMAX', str(memory))
        gdal.SetConfigOption('VSI_CACHE', 'FALSE')
        gdal.SetConfigOption('CHECK_DISK_FREE_SPACE', 'FALSE')
    except Exception as err:
        str(err)
    try:
        cfg.logger.log.debug('option_string: %s' % option_string)
        if int(process_id) == 0:
            progress_gdal = (lambda percentage, m, c: progress_queue.put(
                100 if int(percentage * 100) > 100 else int(percentage * 100),
                False
            ))
        else:
            progress_gdal = None
        to = gdal.TranslateOptions(
            gdal.ParseCommandLine(option_string), callback=progress_gdal
        )
        gdal.Translate(output, input_file, options=to)
    except Exception as err:
        cfg.logger.log.error(str(err))
    cfg.logger.log.debug('end; output: %s' % output)
    logger = cfg.logger.stream.getvalue()
    return False, output, False, logger


# gdal warp processor
def gdal_warp(
        input_file=None, output=None, option_string=None,
        process_parameters=None
):
    process_id = process_parameters[0]
    cfg.temp = process_parameters[1]
    gdal_path = process_parameters[2]
    progress_queue = process_parameters[3]
    memory = process_parameters[4]
    cfg.logger = Log(directory=cfg.temp.dir, multiprocess='0')
    cfg.logger.log.debug('start')
    if gdal_path is not None:
        for d in gdal_path.split(';'):
            try:
                os.add_dll_directory(d)
                cfg.gdal_path = d
            except Exception as err:
                str(err)
    from osgeo import gdal
    # GDAL config
    try:
        gdal.SetConfigOption('GDAL_DISABLE_READDIR_ON_OPEN', 'TRUE')
        gdal.SetConfigOption('GDAL_CACHEMAX', str(memory))
        gdal.SetConfigOption('VSI_CACHE', 'FALSE')
        gdal.SetConfigOption('CHECK_DISK_FREE_SPACE', 'FALSE')
    except Exception as err:
        str(err)
    try:
        cfg.logger.log.debug('option_string: %s' % option_string)
        if int(process_id) == 0:
            progress_gdal = (lambda percentage, m, c: progress_queue.put(
                100 if int(percentage * 100) > 100 else int(percentage * 100),
                False
            ))
        else:
            progress_gdal = None
        to = gdal.WarpOptions(
            gdal.ParseCommandLine(option_string), callback=progress_gdal
        )
        gdal.Warp(output, input_file, options=to)
    except Exception as err:
        cfg.logger.log.error(str(err))
    cfg.logger.log.debug('end; output: %s' % output)
    logger = cfg.logger.stream.getvalue()
    return False, output, False, logger


# convert raster to vector
def raster_to_vector_process(
        raster_path, output_vector_path, field_name=False,
        process_parameters=None,
        connected=4
):
    # process parameters
    process_id = str(process_parameters[0])
    cfg.temp = process_parameters[1]
    gdal_path = process_parameters[2]
    progress_queue = process_parameters[3]
    memory = process_parameters[4]
    if gdal_path is not None:
        for d in gdal_path.split(';'):
            try:
                os.add_dll_directory(d)
                cfg.gdal_path = d
            except Exception as err:
                str(err)
    from osgeo import gdal
    from osgeo import ogr
    from osgeo import osr
    from pathlib import Path
    cfg.logger = Log(directory=cfg.temp.dir, multiprocess=str(process_id))
    cfg.logger.log.debug('start')
    # GDAL config
    try:
        gdal.SetConfigOption('GDAL_DISABLE_READDIR_ON_OPEN', 'TRUE')
        gdal.SetConfigOption('GDAL_CACHEMAX', str(memory))
        gdal.SetConfigOption('VSI_CACHE', 'FALSE')
        gdal.SetConfigOption('CHECK_DISK_FREE_SPACE', 'FALSE')
    except Exception as err:
        str(err)
    # open input with GDAL
    try:
        _r_d = gdal.Open(raster_path, gdal.GA_ReadOnly)
        assert _r_d.RasterXSize
    except Exception as err:
        cfg.logger.log.error(str(err))
        logger = cfg.logger.stream.getvalue()
        return None, str(err), logger
    # create output vector
    d = ogr.GetDriverByName('GPKG')
    cfg.logger.log.debug('output_vector_path: %s' % str(output_vector_path))
    _output_source = d.CreateDataSource(output_vector_path)
    if _output_source is None:
        cfg.logger.log.error('error: %s' % str(output_vector_path))
        logger = cfg.logger.stream.getvalue()
        return None, 'error: %s' % str(output_vector_path), logger
    else:
        sr = osr.SpatialReference()
        cfg.logger.log.debug('sr: %s' % str(sr))
        sr.ImportFromWkt(_r_d.GetProjectionRef())
        p = Path(raster_path)
        name = p.stem
        _output_layer = _output_source.CreateLayer(
            name, sr, ogr.wkbMultiPolygon
        )
        if not field_name:
            field_name = 'DN'
        field_def = ogr.FieldDefn(field_name, ogr.OFTInteger)
        try:
            _output_layer.CreateField(field_def)
        except Exception as err:
            cfg.logger.log.error(str(err))
            logger = cfg.logger.stream.getvalue()
            return None, str(err), logger
        field_index = _output_layer.GetLayerDefn().GetFieldIndex(field_name)
        _raster_band = _r_d.GetRasterBand(1)
        if connected == 8:
            # multipolygon
            opt = ["8CONNECTED=8"]
        else:
            opt = []
        if int(process_id) == 0:
            progress_gdal = (lambda percentage, m, c: progress_queue.put(
                100 if int(percentage * 100) > 100 else int(percentage * 100),
                False
            ))
        else:
            progress_gdal = None
        # raster to polygon
        gdal.Polygonize(
            _raster_band, _raster_band.GetMaskBand(), _output_layer,
            field_index, opt, progress_gdal
        )
        # close raster and vector
        _raster_band = None
        _r_d = None
        _output_layer = None
        _output_source = None
        # open layer
        o_vector = ogr.Open(output_vector_path)
        o_layer = o_vector.GetLayer()
        o_name = o_layer.GetName()
        # get minimum Y
        sql = 'SELECT MIN(miny) FROM "rtree_%s_geom"' % o_name
        sql_output = o_vector.ExecuteSQL(sql, dialect='SQLITE')
        out_feature = sql_output.GetNextFeature()
        min_y = out_feature.GetField(0)
        # release sql results
        o_vector.ReleaseResultSet(sql_output)
        # get maximum Y
        sql = 'SELECT MAX(maxy) FROM "rtree_%s_geom"' % o_name
        sql_output = o_vector.ExecuteSQL(sql, dialect='SQLITE')
        out_feature = sql_output.GetNextFeature()
        max_y = out_feature.GetField(0)
        # release sql results
        o_vector.ReleaseResultSet(sql_output)
        cfg.logger.log.debug('end')
        logger = cfg.logger.stream.getvalue()
        return [output_vector_path, min_y, max_y], False, logger


# raster sieve
def raster_sieve_process(
        process_parameters=None, input_parameters=None, output_parameters=None
):
    # process parameters
    process_id = str(process_parameters[0])
    cfg.temp = process_parameters[1]
    gdal_path = process_parameters[2]
    progress_queue = process_parameters[3]
    memory = process_parameters[4]
    # input_raster parameters
    raster = input_parameters[0]
    sieve_size = input_parameters[1]
    connected = input_parameters[2]
    # output parameters
    output = output_parameters[0]
    data_type = output_parameters[1]
    compress = output_parameters[2]
    compress_format = output_parameters[3]
    output_no_data = output_parameters[4]
    if gdal_path is not None:
        for d in gdal_path.split(';'):
            try:
                os.add_dll_directory(d)
                cfg.gdal_path = d
            except Exception as err:
                str(err)
    from osgeo import gdal
    cfg.logger = Log(directory=cfg.temp.dir, multiprocess=str(process_id))
    cfg.logger.log.debug('start')
    # GDAL config
    try:
        gdal.SetConfigOption('GDAL_DISABLE_READDIR_ON_OPEN', 'TRUE')
        gdal.SetConfigOption('GDAL_CACHEMAX', str(memory))
        gdal.SetConfigOption('VSI_CACHE', 'FALSE')
        gdal.SetConfigOption('CHECK_DISK_FREE_SPACE', 'FALSE')
    except Exception as err:
        str(err)
    # open input with GDAL
    try:
        _r_d = gdal.Open(raster, gdal.GA_ReadOnly)
        cfg.logger.log.debug('raster_path: %s' % raster)
        assert _r_d.RasterXSize
    except Exception as err:
        cfg.logger.log.error(str(err))
        logger = cfg.logger.stream.getvalue()
        return None, str(err), logger
    # create output raster
    out_file = cfg.temp.temporary_raster_path(name_suffix=process_id,
                                              name_prefix=process_id)
    cfg.logger.log.debug('out_file: %s' % str(out_file))
    # start progress
    progress_queue.put(1, False)
    _raster_band = _r_d.GetRasterBand(1)
    offs = _raster_band.GetOffset()
    scl = _raster_band.GetScale()
    if output_no_data is None:
        # nodata value
        output_no_data = _raster_band.GetNoDataValue()
    raster_vector.create_raster_from_reference(
        raster, 1, [out_file], output_no_data, 'GTiff', data_type, compress,
        compress_format, scale=scl, offset=offs
    )
    # raster sieve
    if int(process_id) == 0:
        progress_gdal = (lambda percentage, m, c: progress_queue.put(
            int(percentage * 100), False
        ))
    else:
        progress_gdal = None
    _t_raster = gdal.Open(out_file, gdal.GA_Update)
    _t_band = _t_raster.GetRasterBand(1)
    mask = _raster_band.GetMaskBand()
    gdal.SieveFilter(
        _raster_band, mask, _t_band, sieve_size, connected,
        callback=progress_gdal
    )
    # close rasters
    _raster_band = None
    _r_d = None
    _t_band = None
    _t_raster = None
    if os.path.isfile(out_file):
        try:
            import shutil
            shutil.move(out_file, output)
        except Exception as err:
            cfg.logger.log.error(str(err))
    cfg.logger.log.debug('end')
    # end progress
    progress_queue.put(100, False)
    logger = cfg.logger.stream.getvalue()
    return [[out_file]], False, logger


# convert vector to raster based on the resolution of a raster
def vector_to_raster(
        process_parameters=None, input_parameters=None, output_parameters=None
):
    # process parameters
    process_id = str(process_parameters[0])
    cfg.temp = process_parameters[1]
    gdal_path = process_parameters[2]
    progress_queue = process_parameters[3]
    memory = process_parameters[4]
    # input_raster parameters
    vector_path = input_parameters[0]
    field_name = input_parameters[1]
    reference_raster_path = input_parameters[2]
    nodata_value = input_parameters[3]
    background_value = input_parameters[4]
    burn_values = input_parameters[5]
    x_y_size = input_parameters[6]
    all_touched = input_parameters[7]
    minimum_extent = input_parameters[8]
    # output parameters
    output_path = output_parameters[0]
    output_format = output_parameters[1]
    compress = output_parameters[2]
    compress_format = output_parameters[3]
    if background_value is None:
        background_value = 0
    if nodata_value is None:
        nodata_value = 0
    if output_format is None:
        output_format = 'GTiff'
    if gdal_path is not None:
        for d in gdal_path.split(';'):
            try:
                os.add_dll_directory(d)
                cfg.gdal_path = d
            except Exception as err:
                str(err)
    from osgeo import gdal, ogr
    cfg.logger = Log(directory=cfg.temp.dir, multiprocess=str(process_id))
    cfg.logger.log.debug('start')
    # GDAL config
    try:
        gdal.SetConfigOption('GDAL_DISABLE_READDIR_ON_OPEN', 'TRUE')
        gdal.SetConfigOption('GDAL_CACHEMAX', str(memory))
        gdal.SetConfigOption('VSI_CACHE', 'FALSE')
        gdal.SetConfigOption('CHECK_DISK_FREE_SPACE', 'FALSE')
    except Exception as err:
        str(err)
    # open input with GDAL
    cfg.logger.log.debug('vector_path: %s' % vector_path)
    try:
        _vector = ogr.Open(vector_path)
        _v_layer = _vector.GetLayer()
        # check projection
        proj = _v_layer.GetSpatialRef()
        crs = proj.ExportToWkt()
        crs = crs.replace(' ', '')
        if len(crs) == 0:
            crs = None
        vector_crs = crs
    except Exception as err:
        cfg.logger.log.error(str(err))
        logger = cfg.logger.stream.getvalue()
        return None, str(err), logger
    # create output file
    out_file = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix,
                                            name_prefix=process_id)
    cfg.logger.log.debug('out_file: %s' % str(out_file))
    (gt, reference_crs, unit, xy_count, nd, number_of_bands, block_size,
     scale_offset, data_type) = raster_vector.raster_info(
        reference_raster_path)
    orig_x = gt[0]
    orig_y = gt[3]
    cfg.logger.log.debug('orig_x, orig_y: %s,%s' % (orig_x, orig_y))
    if x_y_size is not None:
        x_size = x_y_size[0]
        y_size = x_y_size[1]
    else:
        x_size = gt[1]
        y_size = abs(gt[5])
    cfg.logger.log.debug('x_size, y_size: %s,%s' % (x_size, y_size))
    # number of x pixels
    grid_columns = int(round(xy_count[0] * gt[1] / x_size))
    # number of y pixels
    grid_rows = int(round(xy_count[1] * abs(gt[5]) / y_size))
    cfg.logger.log.debug('grid_columns, grid_rows: %s,%s'
                         % (grid_columns, grid_rows))
    # check crs
    same_crs = raster_vector.compare_crs(reference_crs, vector_crs)
    cfg.logger.log.debug('same_crs: %s' % str(same_crs))
    if not same_crs:
        cfg.logger.log.error('different crs')
        logger = cfg.logger.stream.getvalue()
        return None, 'different crs', logger
    # calculate minimum extent
    if minimum_extent:
        min_x, max_x, min_y, max_y = _v_layer.GetExtent()
        orig_x = orig_x + x_size * int(round((min_x - orig_x) / x_size))
        orig_y = orig_y + y_size * int(round((max_y - orig_y) / y_size))
        cfg.logger.log.debug('orig_x, orig_y: %s,%s' % (orig_x, orig_y))
        grid_columns = abs(int(round((max_x - min_x) / x_size)))
        grid_rows = abs(int(round((max_y - min_y) / y_size)))
        cfg.logger.log.debug('grid_columns, grid_rows: %s,%s'
                             % (grid_columns, grid_rows))
    driver = gdal.GetDriverByName(output_format)
    temporary_grid = cfg.temp.temporary_raster_path(extension=cfg.tif_suffix)
    # create raster _grid
    _grid = driver.Create(
        temporary_grid, grid_columns, grid_rows, 1, gdal.GDT_Float32,
        options=['COMPRESS=LZW', 'BIGTIFF=YES']
    )
    if _grid is None:
        _grid = driver.Create(
            temporary_grid, grid_columns, grid_rows, 1, gdal.GDT_Int16,
            options=['COMPRESS=LZW', 'BIGTIFF=YES']
        )
    if _grid is None:
        cfg.logger.log.error('error output raster')
        logger = cfg.logger.stream.getvalue()
        return None, 'error grid', logger
    try:
        _grid.GetRasterBand(1)
    except Exception as err:
        cfg.logger.log.error(err)
        logger = cfg.logger.stream.getvalue()
        return None, str(err), logger
    # set raster projection from reference
    _grid.SetGeoTransform([orig_x, x_size, 0, orig_y, 0, -y_size])
    _grid.SetProjection(reference_crs)
    _grid = None
    # start progress
    progress_queue.put(1, False)
    if int(process_id) == 0:
        progress_gdal = (lambda percentage, m, c: progress_queue.put(
            int(percentage * 100), False
        ))
    else:
        progress_gdal = None
    # create output raster
    raster_vector.create_raster_from_reference(
        path=temporary_grid, band_number=1,
        output_raster_list=[out_file],
        nodata_value=nodata_value, driver='GTiff',
        gdal_format='Int32', compress=compress,
        compress_format=compress_format, constant_value=background_value
    )
    # convert reference layer to raster
    _output_raster = gdal.Open(out_file, gdal.GA_Update)
    if all_touched is False or all_touched is None:
        if burn_values is None:
            _o_c = gdal.RasterizeLayer(
                _output_raster, [1], _v_layer, options=[
                    'ATTRIBUTE=%s' % str(field_name), 'COMPRESS=DEFLATE',
                    'PREDICTOR=2', 'ZLEVEL=1'],
                callback=progress_gdal
            )

        else:
            _o_c = gdal.RasterizeLayer(
                _output_raster, [1], _v_layer, burn_values=[burn_values],
                options=['COMPRESS=DEFLATE', 'PREDICTOR=2', 'ZLEVEL=1'],
                callback=progress_gdal
            )
    else:
        if burn_values is None:
            _o_c = gdal.RasterizeLayer(
                _output_raster, [1], _v_layer,
                options=['ATTRIBUTE=%s' % str(field_name),
                         'all_touched=TRUE', 'COMPRESS=DEFLATE',
                         'PREDICTOR=2', 'ZLEVEL=1'], callback=progress_gdal
            )
        else:
            _o_c = gdal.RasterizeLayer(
                _output_raster, [1], _v_layer, burn_values=[burn_values],
                options=['all_touched=TRUE', 'COMPRESS=DEFLATE',
                         'PREDICTOR=2', 'ZLEVEL=1'], callback=progress_gdal
            )
    _output_raster = None
    if os.path.isfile(out_file):
        try:
            import shutil
            shutil.move(out_file, output_path)
        except Exception as err:
            cfg.logger.log.error(str(err))
    cfg.logger.log.debug('end')
    # end progress
    progress_queue.put(100, False)
    logger = cfg.logger.stream.getvalue()
    return [[output_path]], False, logger


# class to create a raster piece from raster sections
class RasterPiece(object):

    def __init__(
            self, section_list, x_min, y_min, x_max, y_max, x_size, y_size,
            y_min_no_boundary=None, y_max_no_boundary=None,
            y_size_boundary_top=None, y_size_boundary_bottom=None,
            y_size_no_boundary=None
    ):
        """
        :param section_list: list of sections
        :param x_min: left position
        :param y_min: top position
        :param x_max: right position
        :param y_max: bottom position
        :param x_size: piece x size
        :param y_size: piece y size
        :param y_min_no_boundary: top position without boundary
        :param y_max_no_boundary: bottom position without boundary
        :param y_size_boundary_top: top pixel buffer
        :param y_size_boundary_bottom: bottom pixel buffer
        :param y_size_no_boundary: piece size without buffer
        """
        self.sections = section_list
        # raster piece size
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        self.x_size = x_size
        self.y_size = y_size
        # raster piece size with boundary
        self.y_min_no_boundary = y_min_no_boundary
        self.y_max_no_boundary = y_max_no_boundary
        # raster size without boundary
        self.y_size_no_boundary = y_size_no_boundary
        # buffer size
        self.y_size_boundary_top = y_size_boundary_top
        self.y_size_boundary_bottom = y_size_boundary_bottom

    # return piece y details as string
    def y_details(self):
        return ('y_min: %s; y_max: %s; y_size: %s; y_min_no_boundary: %s; '
                'y_max_no_boundary: %s; y_size_no_boundary: '
                '%s; y_size_boundary_top: %s; y_size_boundary_bottom: %s'
                % (str(self.y_min), str(self.y_max), str(self.y_size),
                   str(self.y_min_no_boundary),
                   str(self.y_max_no_boundary), str(self.y_size_no_boundary),
                   str(self.y_size_boundary_top),
                   str(self.y_size_boundary_bottom)))


# class to create raster sections
class RasterSection(object):

    def __init__(
            self, x_min, y_min, x_max, y_max, x_size, y_size,
            y_min_no_boundary=None, y_max_no_boundary=None,
            y_size_boundary_top=None, y_size_boundary_bottom=None,
            y_size_no_boundary=None
    ):
        """
        :param x_min: left position
        :param y_min: top position
        :param x_max: right position
        :param y_max: bottom position
        :param x_size: section x size
        :param y_size: section y size
        :param y_min_no_boundary: top position without boundary
        :param y_max_no_boundary: bottom position without boundary
        :param y_size_boundary_top: top pixel buffer
        :param y_size_boundary_bottom: bottom pixel buffer
        :param y_size_no_boundary: section size without buffer
        """
        # section size
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        self.x_size = x_size
        self.y_size = y_size
        # section size with boundary
        self.y_min_no_boundary = y_min_no_boundary
        self.y_max_no_boundary = y_max_no_boundary
        # section size without boundary
        self.y_size_no_boundary = y_size_no_boundary
        # buffer size
        self.y_size_boundary_top = y_size_boundary_top
        self.y_size_boundary_bottom = y_size_boundary_bottom

    # return section y details as string
    def y_details(self):
        return ('y_min: %s; y_max: %s; y_size: %s; y_min_no_boundary: %s; '
                'y_max_no_boundary: %s; y_size_no_boundary: '
                '%s; y_size_boundary_top: %s; y_size_boundary_bottom: %s'
                % (str(self.y_min), str(self.y_max), str(self.y_size),
                   str(self.y_min_no_boundary),
                   str(self.y_max_no_boundary), str(self.y_size_no_boundary),
                   str(self.y_size_boundary_top),
                   str(self.y_size_boundary_bottom)))


# apply multiplicative and additive factors to array
def array_multiplicative_additive_factors(
        array, multiplicative_factor, additive_factor
):
    a = array * float(multiplicative_factor) + float(additive_factor)
    return a



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Band mosaic.

This tool performs the mosaic of single bands, or multiple BandSets.
Corresponding bands in two or more BandSet are merged into a single raster
for each band, covering the extent of input bands.
For instance, this is useful to mosaic several multiband images togheter,
obtaining a multispectral mosaic of bands (e.g., the first band of the mosaic
corresponds to all the first bands of input images).

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> file_list = ['file1_b1.tif', 'file2_b1.tif']
    >>> mosaic_bands = rs.mosaic(input_bands=file_list,output_path='output_directory')
"""  # noqa: E501

from typing import Union, Optional

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.util import files_directories, raster_vector, shared_tools


def mosaic(
        input_bands: Union[list, int, BandSet],
        output_path: Optional[str] = None, overwrite: Optional[bool] = False,
        prefix: Optional[str] = '', nodata_value: Optional[int] = None,
        n_processes: Optional[int] = None,
        available_ram: Optional[int] = None,
        virtual_output: Optional[bool] = False,
        output_name: Optional[str] = None,
        reference_raster_crs: Optional[str] = None,
        bandset_catalog: Optional[BandSetCatalog] = None
) -> OutputManager:
    """Mosaic bands.

    This tool performs the mosaic of corresponding bands from multiple
    BandSets.
    A new raster is created for each band, named as output_name or first band name,
    and followed by band number.

    Args:
        input_bands: list of paths of input rasters, or number of BandSet, or BandSet object.
        output_path: string of output path directory.
        overwrite: if True, output overwrites existing files.
        prefix: optional string for output name prefix.
        nodata_value: value to be considered as nodata.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        virtual_output: if True (and output_path is directory), save output
            as virtual raster of multiprocess parts.
        output_name: string used as general name for all output bands.
        reference_raster_crs: path to a raster to be used as crs reference; if None, the first band is used as reference.
        bandset_catalog: optional type BandSetCatalog for BandSet number.

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = output list

    Examples:
        Perform the mosaic of three BandSets
            >>> # import Remotior Sensus and start the session
            >>> import remotior_sensus
            >>> rs = remotior_sensus.Session()
            >>> catalog = rs.bandset_catalog()
            >>> # create three BandSets
            >>> file_list_1 = ['file1_b1.tif', 'file1_b2.tif', 'file1_b3.tif']
            >>> file_list_2 = ['file2_b1.tif', 'file2_b2.tif', 'file2_b3.tif']
            >>> file_list_3 = ['file3_b1.tif', 'file3_b2.tif', 'file3_b3.tif']
            >>> catalog.create_bandset(file_list_1, bandset_number=1)
            >>> catalog.create_bandset(file_list_2, bandset_number=2)
            >>> catalog.create_bandset(file_list_3, bandset_number=3)
            >>> # start the process
            >>> mosaic_bands = rs.mosaic(input_bands=[1, 2, 3],
            ... output_path='output_directory', bandset_catalog=catalog)
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    cfg.logger.log.debug('input_bands: %s' % str(input_bands))
    if n_processes is None:
        n_processes = cfg.n_processes
    # list of band lists to mosaic
    band_list_list = []
    if len(input_bands) == 1:
        # single BandSet of bands to mosaic
        if type(input_bands[0]) is BandSet or type(
                input_bands[0]
        ) is int:
            # get input list
            band_list = BandSetCatalog.get_band_list(
                input_bands[0], bandset_catalog
            )
            band_list_list.append(band_list)
        else:
            cfg.logger.log.error('band list')
            return OutputManager(check=False)
    else:
        combination_band_list = []
        raster_list = []
        for i in input_bands:
            # list of band sets
            if type(i) is BandSet or type(i) is int:
                # get input list
                band_list = BandSetCatalog.get_band_list(i, bandset_catalog)
                combination_band_list.append(band_list)
            # list of raster paths
            else:
                raster_list.append(i)
        if len(raster_list) > 0:
            band_list_list.append(raster_list)
        elif len(combination_band_list) > 0:
            # combine corresponding bands
            try:
                size = len(combination_band_list[0])
                for s in range(size):
                    new_list = []
                    for c in combination_band_list:
                        new_list.append(c[s])
                    band_list_list.append(new_list)
            except Exception as err:
                cfg.logger.log.error(str(err))
                cfg.messages.error(str(err))
                return OutputManager(check=False)
    # mosaic every list of bands
    n = 0
    output_list = []
    for mosaic_list in band_list_list:
        cfg.logger.log.debug('mosaic_list: %s' % str(mosaic_list))
        cfg.progress.update(
            step=n, steps=len(band_list_list), minimum=10, maximum=99,
            message='processing list %s' % str(n + 1),
            percentage=n / len(band_list_list)
        )
        # list of inputs
        prepared = shared_tools.prepare_input_list(
            mosaic_list, reference_raster_crs, n_processes=n_processes
        )
        input_raster_list = prepared['input_list']
        raster_info = prepared['information_list']
        nodata_list = prepared['nodata_list']
        warped = prepared['warped']
        try:
            output_data_type = raster_info[0][8]
        except Exception as err:
            str(err)
            output_data_type = None
        if nodata_value is None:
            nodata_value = nodata_list[0]
        if warped:
            virtual_output = False
        # check output path
        if output_name is None:
            p = shared_tools.join_path(
                output_path, '{}{}'.format(
                    prefix, files_directories.file_name(input_raster_list[0])
                )
            ).replace('\\', '/')
        else:
            p = shared_tools.join_path(
                output_path, '{}{}{}'.format(prefix, output_name, n)
            ).replace('\\', '/')
        out_path, vrt_r = files_directories.raster_output_path(
            p, overwrite=overwrite
        )
        output_list.append(out_path)
        if virtual_output:
            try:
                # create virtual raster
                raster_vector.create_virtual_raster_2_mosaic(
                    input_raster_list=input_raster_list, output=out_path,
                    dst_nodata=nodata_value, data_type=output_data_type
                )
            except Exception as err:
                cfg.logger.log.error(str(err))
                cfg.messages.error(str(err))
                return OutputManager(check=False)
        else:
            vrt_file = cfg.temp.temporary_raster_path(extension=cfg.vrt_suffix)
            try:
                # create virtual raster
                raster_vector.create_virtual_raster_2_mosaic(
                    input_raster_list=input_raster_list, output=vrt_file,
                    dst_nodata=nodata_value, data_type=output_data_type
                )
                # copy raster
                cfg.multiprocess.gdal_copy_raster(
                    vrt_file, out_path, 'GTiff', cfg.raster_compression, 'LZW',
                    additional_params='-ot %s' % str(output_data_type),
                    n_processes=n_processes, available_ram=available_ram
                )
            except Exception as err:
                cfg.logger.log.error(str(err))
                cfg.messages.error(str(err))
                return OutputManager(check=False)
        n = n + 1
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; mosaic: %s' % str(output_list))
    return OutputManager(paths=output_list)



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Band combination.

This tool is intended for combining classifications in order to get a
raster where each value corresponds to a combination of class values.
A unique value is assigned to each combination of values.
The output is a raster made of unique values corresponding to combinations
of values.
An output text file describes the correspondence between unique values
and combinations, as well as the statistics of each combination.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> combination = rs.band_combination(input_bands=['path_1', 'path_2'],
    ... output_path='output_path')
"""

import io
from typing import Union, Optional

import numpy as np

from remotior_sensus.core import configurations as cfg, table_manager as tm
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.core.processor_functions import (
    cross_rasters, raster_unique_values
)
from remotior_sensus.util import (
    shared_tools, files_directories, raster_vector, read_write_files
)


def band_combination(
        input_bands: Union[list, int, BandSet],
        output_path: Optional[str] = None, nodata_value: Optional[int] = None,
        overwrite: Optional[bool] = False,
        n_processes: Optional[int] = None, available_ram: Optional[int] = None,
        bandset_catalog: Optional[BandSetCatalog] = None,
        extent_list: Optional[list] = None,
        column_name_list: Optional[list] = None,
        output_table: Optional[bool] = True,
        progress_message: Optional[bool] = True
) -> OutputManager:
    """Calculation of band combination.

    This tool allows for the combination of rasters or bands loaded in a
    BandSet.
    This tool is intended for combining classifications in order to get a
    raster where each value corresponds to a combination of class values.
    Input raster values must be integer type.
    The output is a combination raster and a text file reporting the
    statistics of each combination.

    Args:
        input_bands: list of paths of input rasters, or number of BandSet, or BandSet object.
        output_path: path of the output raster.
        overwrite: if True, output overwrites existing files.
        nodata_value: input value to be considered as nodata.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: BandSetCatalog object required if input_bands is a BandSet number.
        extent_list: list of boundary coordinates left top right bottom.
        column_name_list: list of strings corresponding to input bands used 
            as column names in output table, if None then column names are extracted for input band names.
        output_table: if True then calculate output table; 
            if False then calculate only array of combinations and sum.
        progress_message: if True then start progress message; 
            if False does not start the progress message (useful if launched from other tools).

    Returns:
        If output_table is True returns the :func:`~remotior_sensus.core.output_manager.OutputManager` object with
            - paths = [output raster path, output table path]
        
        If output_table is False returns the :func:`~remotior_sensus.core.output_manager.OutputManager` object with
            - paths = [virtual raster path]
            - extra = {'combinations': array of combinations, 'sums': array of the sums of values}

    Examples:
        Combination using two rasters having paths path_1 and path_2
            >>> combination = band_combination(input_bands=['path_1', 'path_2'],output_path='output_path')
            
        The combination raster and the table are finally created; the paths can be retrieved from the output that is an :func:`~remotior_sensus.core.output_manager.OutputManager` object
            >>> raster_path, table_path = combination.paths
            >>> print(raster_path)
            output_path

        Combination using a virtual raster as output file
            >>> combination = band_combination(input_bands=['path_1', 'path_2'],output_path='output_path.vrt')
            >>> raster_path, table_path = combination.paths
            >>> print(raster_path)
            output_path.vrt

        Using input BandSet number
            >>> catalog = BandSetCatalog()
            >>> combination = band_combination(input_bands=1,output_path='output_path',bandset_catalog=catalog)

        Using input BandSet
            >>> catalog = BandSetCatalog()
            >>> combination = band_combination(input_bands=catalog.get_bandset(1),output_path='output_path')
    """  # noqa: E501
    if progress_message:
        cfg.logger.log.info('start')
        cfg.progress.update(
            process=__name__.split('.')[-1].replace('_', ' '),
            message='starting', start=True
        )
    # prepare process files
    prepared = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path, overwrite=overwrite,
        n_processes=n_processes, bandset_catalog=bandset_catalog,
        box_coordinate_list=extent_list
    )
    input_raster_list = prepared['input_raster_list']
    name_list = prepared['name_list']
    out_path = prepared['output_path']
    vrt_r = prepared['virtual_output']
    vrt_path = prepared['temporary_virtual_raster']
    n_processes = prepared['n_processes']
    # dummy bands for memory calculation as the number of input raster
    dummy_bands = len(input_raster_list) + 1
    cfg.multiprocess.run(
        raster_path=vrt_path, function=raster_unique_values,
        keep_output_argument=True, n_processes=n_processes,
        available_ram=available_ram, dummy_bands=dummy_bands,
        progress_message='unique values', min_progress=2, max_progress=50
    )
    cfg.multiprocess.multiprocess_unique_values()
    cmb = cfg.multiprocess.output
    cfg.logger.log.debug('len(cmb): %s; cmb[0]: %s' % (len(cmb), str(cmb[0])))
    # random variable list
    rnd_var_list = None
    # array of combinations
    dtype_list = []
    for c in range(len(input_raster_list)):
        dtype_list.append(('f%s' % str(c), 'int64'))
    cmb_arr = np.rec.fromarrays(np.asarray(cmb).T, dtype=dtype_list)
    cmb_arr_ravel = np.asarray(cmb).ravel()
    max_v = np.nanmax(cmb_arr_ravel)
    if np.sum(cmb_arr_ravel <= 0) < 1:
        # calculation data type
        calc_data_type = np.uint32
        output_data_type = cfg.uint32_dt
        calc_nodata = cfg.nodata_val_UInt32
        add_c = 1
    else:
        add_c = 1 - np.nanmin(cmb_arr_ravel)
        calc_data_type = np.int32
        output_data_type = cfg.int32_dt
        calc_nodata = cfg.nodata_val_Int32
    # expression builder
    max_dig = 10 - len(str(np.sum(max_v)))
    if max_dig < 0:
        max_dig = 2
    rec_combinations_array = None
    reclassification_list = None
    t = 0
    method = 1
    while t < 5000:
        t += 1
        rnd_var_list = []
        expression_comb = []
        for y in range(len(input_raster_list)):
            if method == 1:
                exp_r = int(np.random.random() * 10) + 1
                if exp_r > max_dig:
                    exp_r = max_dig
                const_v = int(10 ** exp_r)
                method += 1
            elif method == 2:
                const_v = int(10 ** max_dig)
                method += 1
            else:
                exp_r = int(np.random.random() * 10) + 1
                if exp_r > 8:
                    exp_r = 8
                const_v = int(10 ** exp_r)
                method = 1
            if const_v < 1:
                const_v = 3
            rnd_var = int(const_v * np.random.random())
            if rnd_var == 0:
                rnd_var = 1
            # avoid too large numbers
            while np.sum(
                    rnd_var * (np.array(max_v, dtype=np.float32) + add_c)
            ) > calc_nodata:
                rnd_var = int(rnd_var / 2)
            rnd_var_list.append(rnd_var)
            # expression combination
            expression_comb.append(
                '("f%s" + %s) * %s' % (str(y), str(add_c), str(rnd_var))
            )
            expression_comb.append(' + ')
        expression_comb.pop(-1)
        joined_expression_comb = ''.join(expression_comb)
        rec_combinations_array = tm.calculate(
            matrix=cmb_arr, expression_string=joined_expression_comb,
            output_field_name='id', progress_message=False
        )
        new_val = list(range(1, rec_combinations_array.shape[0] + 1))
        rec_combinations_array = tm.sort_table_by_field(
            rec_combinations_array, 'id'
        )
        rec_combinations_array = tm.append_field(
            rec_combinations_array, 'new_val', new_val, 'int64'
        )
        # check if unique new values are the same number as combinations
        uni = np.unique(rec_combinations_array.id)
        if uni.shape == cmb_arr.shape:
            reclassification_list = sorted(list(uni))
            break
    expression = []
    for r in range(len(rnd_var_list)):
        expression.append(
            '(%s[::, ::, %s] + %s) * %s' % (
                cfg.array_function_placeholder, str(r), str(add_c),
                str(rnd_var_list[r]))
        )
        expression.append(' + ')
    expression.pop(-1)
    joined_expression = ''.join(expression)
    cfg.logger.log.debug('joined_expression: %s' % joined_expression)
    # dummy bands for memory calculation
    dummy_bands = 3
    # combination calculation
    cfg.multiprocess.run(
        raster_path=vrt_path, function=cross_rasters,
        function_argument=reclassification_list,
        function_variable=joined_expression, dummy_bands=dummy_bands,
        calculation_datatype=calc_data_type, use_value_as_nodata=nodata_value,
        any_nodata_mask=True, output_raster_path=out_path,
        output_data_type=output_data_type, output_nodata_value=calc_nodata,
        compress=cfg.raster_compression, n_processes=n_processes,
        available_ram=available_ram, keep_output_argument=True,
        virtual_raster=vrt_r, progress_message='cross rasters',
        min_progress=50, max_progress=90
    )
    cfg.progress.update(message='output table', step=90)
    # calculate sum of values
    cfg.multiprocess.multiprocess_sum_array(nodata_value)
    sum_val = cfg.multiprocess.output
    if not output_table:
        return OutputManager(
            paths=[vrt_path], extra={
                'combinations': rec_combinations_array, 'sums': sum_val
            }
        )
    else:
        # get pixel unit
        (gt, crs, un, xy_count, nd, number_of_bands, block_size, scale_offset,
         data_type) = raster_vector.raster_info(out_path)
        p_x = gt[1]
        p_y = abs(gt[5])
        joined_table = tm.join_tables(
            table1=rec_combinations_array, table2=sum_val,
            field1_name='new_val', field2_name='new_val',
            nodata_value=cfg.nodata_val_Int64, join_type='left',
            progress_message=False, min_progress=90, max_progress=100
        )
        # name of combination field in output table
        combination_name_list = []
        p = 0
        for r_name in name_list:
            if column_name_list is None:
                combination_name_list.append(r_name[0:16])
            else:
                try:
                    combination_name_list.append(column_name_list[p])
                except Exception as err:
                    str(err)
                    combination_name_list.append(r_name[0:16])
                p += 1
        # create table
        table = _combination_table(
            joined_table[joined_table['sum'] != cfg.nodata_val_Int64],
            combination_name_list, un, p_x, p_y
        )
        # save combination to table
        tbl_out = shared_tools.join_path(
            files_directories.parent_directory(out_path), '{}{}'.format(
                files_directories.file_name(
                    out_path, suffix=False
                ), cfg.csv_suffix
            )
        ).replace('\\', '/')
        read_write_files.write_file(table, tbl_out)
        cfg.progress.update(end=True)
        cfg.logger.log.info(
            'end; band combination: %s; table: %s'
            % (str(out_path), str(tbl_out))
        )
        return OutputManager(
            paths=[out_path, tbl_out], extra={
                'combinations': rec_combinations_array,
                'sums': sum_val
            }
        )


#
def _combination_table(
        table: np.array, combination_names: list, crs_unit: str,
        pixel_size_x: int, pixel_size_y: int
) -> str:
    """Creation of combination table.

    Creates text for table where each value corresponds to a combination 
    of class values, and the area statistics of each combination,
    if the input bands are in cartographic coordinates.

    Args:
        table: table of band combination.
        combination_names: list of names of combinations.
        crs_unit: unit of crs used for area calculation.
        pixel_size_x: pixel size along x.
        pixel_size_y: pixel size along y.
        
    Returns:
        The text of the combination table.
    """
    text = []
    cv = cfg.comma_delimiter
    nl = cfg.new_line
    # table
    if 'degree' not in crs_unit:
        output_field_names = ['RasterValue']
        input_field_names = ['new_val']
        for c in range(len(combination_names)):
            output_field_names.append(combination_names[c])
            input_field_names.append('f%s' % str(c))
        output_field_names.append('PixelSum')
        output_field_names.append('Area [%s^2]' % crs_unit)
        input_field_names.append('sum')
        input_field_names.append('area')
        cross_class = tm.calculate(
            matrix=table, expression_string='"sum" * %s * %s' % (
                str(pixel_size_x), str(pixel_size_y)),
            output_field_name='area', progress_message=False
        )
    else:
        output_field_names = ['RasterValue']
        input_field_names = ['new_val']
        for c in range(len(combination_names)):
            output_field_names.append(combination_names[c])
            input_field_names.append('f%s' % str(c))
        output_field_names.append('PixelSum')
        input_field_names.append('sum')
        cross_class = table
    redefined = tm.redefine_matrix_columns(
        matrix=cross_class, input_field_names=input_field_names,
        output_field_names=output_field_names, progress_message=False
    )
    # create stream handler
    stream1 = io.StringIO()
    np.savetxt(stream1, redefined, delimiter=cv, fmt='%1.2f')
    matrix_value = stream1.getvalue()
    for c in output_field_names:
        text.append(c)
        text.append(cv)
    text.pop(-1)
    text.append(nl)
    text.append(matrix_value.replace('.00', ''))
    text.append(nl)
    joined_text = ''.join(text)
    return joined_text



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""
Band resample.

This tool allows for resampling and reprojecting bands.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> rs.band_resample(input_bands=['path_1', 'path_2'],
    ... output_path='output_path')
"""  # noqa: E501

from typing import Union, Optional

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.util import shared_tools, raster_vector


def band_resample(
        input_bands: Union[list, int, BandSet],
        output_path: Optional[str] = None,
        epsg_code: Optional[str] = None,
        align_raster: Optional[Union[str, BandSet, int]] = None,
        overwrite: Optional[bool] = False,
        resampling: Optional[str] = None,
        nodata_value: Optional[int] = None,
        x_y_resolution: Optional[Union[list, int]] = None,
        resample_pixel_factor: Optional[float] = None,
        output_data_type: Optional[str] = None,
        same_extent: Optional[bool] = False,
        virtual_output: Optional[bool] = None,
        compress=None, compress_format=None,
        prefix: Optional[str] = '',
        extent_list: Optional[list] = None,
        n_processes: Optional[int] = None,
        available_ram: Optional[int] = None,
        bandset_catalog: Optional[BandSetCatalog] = None
):
    """Performs band resample and reprojection.

    This tool performs the resampling and reprojection of raster bands. 
    Available resampling methods are:

        - nearest_neighbour
        - average
        - sum
        - maximum
        - minimum
        - mode
        - median
        - first_quartile
        - third_quartile

    Args:
        input_bands: input of type BandSet or list of paths or integer
            number of BandSet.
        output_path: string of output path directory or list of paths.
        epsg_code: optional EPSG code for output.
        align_raster: string path of raster used for aligning output pixels and projections.
        overwrite: if True, output overwrites existing files.
        resampling: method of resample such as nearest_neighbour (default), average, sum, maximum, minimum, mode, median, first_quartile, third_quartile.
        nodata_value: value to be considered as nodata.
        x_y_resolution: integer pixel size of output raster or pixel size as list of x, y.
        resample_pixel_factor: define output resolution by multiplying original pixel size to this value.
        output_data_type: optional raster output data type, if None the data type is the same as input raster.
        same_extent: if True, output extent is the same as align_raster.
        virtual_output: if True (and output_path is directory), save output as virtual raster of multiprocess parts.
        prefix: optional string for output name prefix.
        extent_list: list of boundary coordinates left top right bottom.
        compress: if True, compress output.
        compress_format: format of compressions such as LZW or DEFLATE.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: optional type BandSetCatalog for BandSet number.

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = output list

    Examples:
        Perform the band resample
            >>> band_resample(input_bands=['path_1', 'path_2'],
            ... output_path='output_path', resampling='mode',
            ... resample_pixel_factor=2)
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    if resampling == 'nearest_neighbour':
        resample = 'near'
    elif resampling == 'average':
        resample = 'average'
    elif resampling == 'sum':
        resample = 'sum'
    elif resampling == 'maximum':
        resample = 'max'
    elif resampling == 'minimum':
        resample = 'min'
    elif resampling == 'mode':
        resample = 'mode'
    elif resampling == 'median':
        resample = 'med'
    elif resampling == 'first_quartile':
        resample = 'q1'
    elif resampling == 'third_quartile':
        resample = 'q3'
    else:
        resample = 'near'
    # prepare process files
    prepared = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path, overwrite=overwrite,
        n_processes=n_processes, box_coordinate_list=extent_list,
        bandset_catalog=bandset_catalog, prefix=prefix,
        multiple_output=True, multiple_input=True,
        virtual_output=virtual_output
    )
    input_raster_list = prepared['input_raster_list']
    raster_info = prepared['raster_info']
    n_processes = prepared['n_processes']
    nodata_list = prepared['nodata_list']
    output_list = prepared['output_list']
    if x_y_resolution is not list:
        x_y_resolution = [x_y_resolution, x_y_resolution]
    if resample_pixel_factor is None:
        resample_pixel_factor = 1
    try:
        resample_pixel_factor = float(resample_pixel_factor)
    except Exception as err:
        cfg.logger.log.error(str(err))
        cfg.messages.error(str(err))
        return OutputManager(check=False)
    if epsg_code is None:
        epsg = False
    else:
        epsg = None
    resample_parameters = None
    if align_raster is None:
        left = top = right = bottom = p_x = p_y = align_sys_ref = None
    else:
        # raster extent and pixel size
        (gt, crs, crs_unit, xy_count, nd, number_of_bands, block_size,
         scale_offset, data_type) = raster_vector.raster_info(align_raster)
        # copy raster
        left = gt[0]
        top = gt[3]
        right = gt[0] + gt[1] * xy_count[0]
        bottom = gt[3] + gt[5] * xy_count[1]
        p_x = gt[1]
        p_y = abs(gt[5])
        # check projections
        align_sys_ref = raster_vector.auto_set_epsg(align_raster)
        epsg = False
        new_p_x = p_x * resample_pixel_factor
        new_p_y = p_y * resample_pixel_factor
        resample_parameters = '-tr %s %s -te %s %s %s %s' % (
            str(new_p_x), str(new_p_y), str(left), str(bottom), str(right),
            str(top))
    if compress_format == 'DEFLATE21':
        compress_format = 'DEFLATE -co PREDICTOR=2 -co ZLEVEL=1'
    min_progress = 1
    one_progress = int((99 - 1) / len(input_raster_list))
    max_progress = one_progress
    for band in range(0, len(input_raster_list)):
        # raster extent and pixel size
        (left_input, top_input, right_input, bottom_input, p_x_input,
         p_y_input, proj_input,
         unit_input) = raster_vector.image_geotransformation(
            input_raster_list[band]
        )
        if output_data_type is None:
            output_data_type = raster_info[band][8]
        # calculate minimal extent
        if align_raster is not None:
            input_sys_ref = raster_vector.get_spatial_reference(proj_input)
            left_projected, top_projected = (
                raster_vector.project_point_coordinates(
                    left_input, top_input, input_sys_ref, align_sys_ref
                )
            )
            right_projected, bottom_projected = (
                raster_vector.project_point_coordinates(
                    right_input, bottom_input, input_sys_ref, align_sys_ref
                )
            )
            if same_extent is False:
                # minimum extent
                if left_projected < left:
                    left_output = left - int(
                        2 + (left - left_projected) / p_x
                    ) * p_x
                else:
                    left_output = left + int(
                        (left_projected - left) / p_x - 2
                    ) * p_x
                if right_projected > right:
                    right_output = right + int(
                        2 + (right_projected - right) / p_x
                    ) * p_x
                else:
                    right_output = right - int(
                        (right - right_projected) / p_x - 2
                    ) * p_x
                if top_projected > top:
                    top_output = top + int(
                        2 + (top_projected - top) / p_y
                    ) * p_y
                else:
                    top_output = top - int(
                        (top - top_projected) / p_y - 2
                    ) * p_y
                if bottom_projected > bottom:
                    bottom_output = bottom + int(
                        (bottom_projected - bottom) / p_y - 2
                    ) * p_y
                else:
                    bottom_output = bottom - int(
                        2 + (bottom - bottom_projected) / p_y
                    ) * p_y
            else:
                left_output = left
                top_output = top
                right_output = right
                bottom_output = bottom
            resample_parameters = '-tr %s %s -te %s %s %s %s ' % (
                str(p_x), str(p_y), str(left_output), str(bottom_output),
                str(right_output), str(top_output))
        # use epsg
        elif epsg_code is not None:
            # spatial reference
            resample_parameters = None
            try:
                epsg = int(epsg_code)
            except Exception as err:
                cfg.logger.log.error(str(err))
                cfg.messages.error(str(err))
                return OutputManager(check=False)
            if same_extent is False:
                try:
                    resample_parameters = '-tr %s %s' % (
                        str(float(x_y_resolution[0])),
                        str(float(x_y_resolution[1])))
                except Exception as err:
                    str(err)
            else:
                left_output = left
                top_output = top
                right_output = right
                bottom_output = bottom
                try:
                    resample_parameters = '-tr %s %s -te %s %s %s %s ' % (
                        str(float(x_y_resolution[0])),
                        str(float(x_y_resolution[1])),
                        str(left_output), str(bottom_output),
                        str(right_output), str(top_output)
                    )
                except Exception as err:
                    str(err)
        # resample
        else:
            if epsg is False:
                epsg = None
            p_x = p_x_input * resample_pixel_factor
            p_y = p_y_input * resample_pixel_factor
            if same_extent is False:
                try:
                    resample_parameters = '-tr %s %s' % (str(p_x), str(p_y))
                except Exception as err:
                    str(err)
            else:
                left_output = left
                top_output = top
                right_output = right
                bottom_output = bottom
                try:
                    resample_parameters = '-tr %s %s -te %s %s %s %s ' % (
                        str(p_x), str(p_y), str(left_output),
                        str(bottom_output), str(right_output), str(top_output)
                    )
                except Exception as err:
                    str(err)
        if epsg is not None:
            if epsg is False:
                epsg_output = proj_input
            else:
                epsg_output = 'epsg:%s' % str(epsg)
        else:
            epsg_output = None
        if nodata_value is None:
            nodata_value = nodata_list[band]
        cfg.logger.log.debug('resample_parameters: %s' % resample_parameters)
        # calculation
        raster_vector.gdal_warping(
            input_raster=input_raster_list[band],
            output=output_list[band],
            output_format='GTiff',
            resample_method=resample,
            t_srs=epsg_output,
            compression=compress,
            compress_format=compress_format,
            additional_params=resample_parameters,
            raster_data_type=output_data_type,
            dst_nodata=nodata_value,
            available_ram=available_ram,
            n_processes=n_processes,
            min_progress=min_progress, max_progress=max_progress
        )
        min_progress = max_progress
        max_progress += one_progress
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; band resample: %s' % str(output_list))
    return OutputManager(paths=output_list)



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import files_directories


class TestMosaicBands(TestCase):

    def test_mosaic_bands(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
            )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        catalog = rs.bandset_catalog()
        file_list = ['./data/S2_2020-01-01/S2_B02.tif',
                     './data/S2_2020-01-02/S2_B02.tif']
        temp = cfg.temp.dir
        cfg.logger.log.debug('>>> test mosaic')
        mosaic = rs.mosaic(file_list, temp)
        self.assertTrue(files_directories.is_file(mosaic.paths[0]))
        self.assertTrue(mosaic.check)
        file_list_1 = ['S2_2020-01-01/S2_B02.tif', 'S2_2020-01-01/S2_B03.tif',
                       'S2_2020-01-01/S2_B04.tif']
        file_list_2 = ['S2_2020-01-02/S2_B02.tif', 'S2_2020-01-02/S2_B03.tif',
                       'S2_2020-01-02/S2_B04.tif']
        file_list_3 = ['S2_2020-01-03/S2_B02.tif', 'S2_2020-01-03/S2_B03.tif',
                       'S2_2020-01-03/S2_B04.tif']
        root_directory = './data'
        catalog.create_bandset(
            file_list_1, wavelengths=['Sentinel-2'], bandset_number=1,
            root_directory=root_directory
            )
        catalog.create_bandset(
            file_list_2, wavelengths=['Sentinel-2'], bandset_number=2,
            root_directory=root_directory
            )
        catalog.create_bandset(
            file_list_3, wavelengths=['Sentinel-2'], bandset_number=3,
            root_directory=root_directory
            )
        bandset_list = [catalog.get_bandset(1), catalog.get_bandset(2),
                        catalog.get_bandset(3)]
        cfg.logger.log.debug('>>> test mosaic BandSet')
        mosaic = rs.mosaic(bandset_list, temp)
        self.assertTrue(files_directories.is_file(mosaic.paths[0]))
        self.assertTrue(mosaic.check)
        bandset_list = [1, 2]
        mosaic = rs.mosaic(
            bandset_list, output_path=temp, bandset_catalog=catalog
            )
        self.assertTrue(files_directories.is_file(mosaic.paths[0]))
        self.assertTrue(mosaic.check)

        # clear temporary directory
        rs.close()
